"""
The `base` sub-package contains three modules:

`html_creator`
--------------

A basic HTML template creator for Orthology's full report.

`plotter`
---------
Contains all functions used to generate plots for TriFusion's
Statistics screen and for TriStats.

`sanity`
--------
Contains several sanity checks performed during the execution
of the TriSeq CLI program.
"""
